/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <string.h>
#include <dlfcn.h>
#include <search.h>
#include <unistd.h>

#include "iomanager.h"
#include "debug.h"
#include "table.h"

#define ADD_LIST(x,h) x->next=h; h=x;

typedef struct{
    void *dlhandle;
    ioman_create_func *create_device;
    ioman_register_func *add_register;
    ioman_interrupt_func *add_interrupt;
    ioman_clock_func *add_clock;
    ioman_param_func *set_param;
    ioman_uses_func *set_device;
}device_funcs;

typedef struct param_entry {
    struct param_entry* next;
    char* device;
    char* key;
    char* value;
} param_entry;

static char *so_paths[]={
    "./devices/%s."SOEXT,   /* look relative to self (build directory) */
    DEVICEDIR"/%s."SOEXT,   /* or here (installed directory) */
    "%s."SOEXT,             /* as a last stab, library search path */
    NULL
};

static char *conf_paths[]={
    "%s",
    "config/%s",
    CONFDIR"/%s",
    NULL
};

static char *run_paths[]={
    "%s",
    "run/%s",
    RUNDIR"/%s",
    NULL
};

/* white space definition for string tokenizing */
static const char wspace[]=" \n\t";

#define PARSER_MAX_REGS 256
#define PARSER_MAX_ARGS 16
#define PARSER_LINE_BUFF_SIZE 1024

static char *line_buff=NULL;

typedef struct{
    uint8_t reg;
    char *name;
}io_reg;
static table_t *io_reg_table=NULL;

/* open the config files, using reasonable paths */
static FILE *open_config(char *config){
    FILE *conffile=NULL;
    char fname[256];
    int i=0;

    while(conffile == NULL && conf_paths[i] != NULL){
        snprintf(fname, sizeof(fname), conf_paths[i], config);
        fname[255]='\0';
        
        conffile = fopen(fname, "r");
        i++;
    }

    return conffile;
}

/* open the run files, using reasonable paths */
static FILE *open_run(char *run){
    FILE *runfile=NULL;
    char fname[256];
    int i=0;

    while(runfile == NULL && run_paths[i] != NULL){
        snprintf(fname, sizeof(fname), run_paths[i], run);
        fname[255]='\0';
        
        runfile = fopen(fname, "r");
        i++;
    }

    return runfile;
}

static void regfree(void * r){
    free(((io_reg*)r)->name);
    free(r);
}

static int regcompare(const void * a, const void *b){
    return strcasecmp(((io_reg*)a)->name, ((io_reg*)b)->name);
}

static param_entry* copy_param(const char* device, const char* key, const char* value) {
    param_entry* retval;
    int devlen = device ? strlen(device) : 0;
    int keylen = key ? strlen(key) : 0;
    int vallen = value ? strlen(value) : 0;
    char *buffer;

    /* We take the structure and string space in a single malloc(), this
     * means you only free() the structure, and don't have to worry about
     * freeing the internal string pointers.  Makes this code a little
     * messier, but makes the freeing code much simplier and less errorprone.
     * Also, we have less malloc()'s to worry about failing. */

    retval = (param_entry*)malloc(sizeof(param_entry)
            + devlen + keylen + vallen + 3);
    if(retval == NULL) {
        return NULL;
    }
    buffer = ((char*)retval) + sizeof(param_entry);

    retval->next = NULL;
    if(device) {
        retval->device = buffer;
        strcpy(buffer, device);
        buffer += devlen + 1;
    } else {
        retval->device = NULL;
    }

    if(key) {
        retval->key = buffer;
        strcpy(buffer, key);
        buffer += keylen + 1;
    } else {
        retval->key = NULL;
    }

    if(value) {
        retval->value = buffer;
        strcpy(buffer, value);
    } else {
        retval->value = NULL;
    }

    return retval;
}

static int parse_line(FILE *conf, int *lineno, int *argc, char *argv[PARSER_MAX_ARGS]){
    char *nextarg;
    *argc=0;
    
    /* loop until we hit the end or find a directive */
    while(*argc==0 && fgets(line_buff, PARSER_LINE_BUFF_SIZE, conf)!=NULL){
        (*lineno)++;
        argv[0]=strtok(line_buff, wspace);
        
        if(argv[0] == NULL || argv[0][0]==';' || argv[0][0]=='#'){
            /* ignore blank lines and comments */
            continue;
        }
        (*argc)++;

        /* loop until max args or eol or comment */
        while(*argc<PARSER_MAX_ARGS &&
                (nextarg=strtok(NULL, wspace))!=NULL && nextarg[0]!=';' && nextarg[0]!='#'){
            argv[*argc]=nextarg;
            (*argc)++;
        }
    }

    return *argc;
}

static error_t set_interrupt(interrupt_obj *interrupt) {
    SET_FLAG(interrupt->chip->interrupts, interrupt->vector_mask);
    return 0;
}

static error_t clear_interrupt(interrupt_obj *interrupt) {
    CLR_FLAG(interrupt->chip->interrupts, interrupt->vector_mask);
    return 0;
}

static error_t device_add_clock(device_obj *device, clock_obj *clock) {
    /* Make sure not already in list */
    clock_obj *head = device->chip->clock_head;
    while(head) {
        if(head == clock) {
            return 0;
        }
        head = head->next;
    }

    clock->next = NULL;
    ADD_LIST(clock, device->chip->clock_head);
    return 0;
}

static error_t device_remove_clock(device_obj *device, clock_obj *clock) {
    /* Look for it in the list */
    clock_obj *head = device->chip->clock_head, *prev = NULL;
    while(head) {
        if(head == clock) {
            /* Found it */
            if(prev == NULL) {
                device->chip->clock_head = head->next;
            } else {
                prev->next = head->next;
            }
            head->next = NULL;
            return 0;
        }
        prev = head;
        head = head->next;
    }
    return 0;
}

static error_t io_reg_table_lookup(char *reg, uint8_t *regnum){
    /* look up cpu code in hash table, return register number */
    io_reg *result;
    io_reg query;
    
    query.name=reg;
    
    result=(io_reg*)table_find(io_reg_table, &query);
    
    if(result==NULL){
        printf("Register %s not found.\n", reg);
        return -1;
    }
    
    /* calculate offset into the table, that is register number */
    *regnum=result->reg;
    
    return 0;
}

error_t atmega_parse_runfile(atmega_chip ***chips, int *nchips, char *runfile){
    int lineno=0;
    int argc;
    char *argv[PARSER_MAX_REGS];
    char buff[1024];
    FILE *runf;
    char *directive;
    int node=0;
    int chipid=-1;
    char *config=NULL, *program=NULL;
    param_entry *param_list = NULL, *thisparam, *nextparam;
    device_obj *thisdev;
    
    line_buff=buff;

    if((runf=open_run(runfile))==NULL){
        printf("Cannot open runfile %s.\n", runfile);
        return -1;
    }

    *nchips=0;

    while(parse_line(runf, &lineno, &argc, argv)){
        directive=argv[0];
        
        if(strcasecmp(directive, "NCHIPS")==0){
            if(argc!=2){
                printf("Invalid NCHIPS directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            *nchips=strtol(argv[1], NULL, 0);

            if(*nchips < 1){
                printf("NCHIPS must be >= 1, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            *chips=malloc(sizeof(atmega_chip*) * (*nchips));

            if(*chips == NULL){
                printf("Unable to allocate memory for %d chips, line %d.\n", *nchips, lineno);
                fclose(runf);
                return -1;
            }

            printf( "Number of chips: %d\n", *nchips );
            fflush( stdout );
        }else if(strcasecmp(directive, "NODE")==0){
            if(node){
                printf("You must end a NODE section before beginning another, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            if(argc!=1){
                printf("Invalid NODE directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            node=1;
            chipid++;

            if(chipid >= *nchips){
                printf("Too many NODE sections; Chip %d at line %d, NCHIPS was %d.\n", chipid, lineno, *nchips);
                fclose(runf);
                return -1;
            }
        }else if(strcasecmp(directive, "CONFIG")==0){
            if(!node){
                printf("CONFIG directives are only valid in a NODE section, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            if(argc != 2){
                printf("malformed CONFIG directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            if(config != NULL){
                printf("Duplicate CONFIG directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            config=strdup(argv[1]);
        }else if(strcasecmp(directive, "PROGRAM")==0){
            if(!node){
                printf("PROGRAM directives are only valid in a NODE section, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            if(argc != 2){
                printf("malformed PROGRAM directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            if(program != NULL){
                printf("Duplicate PROGRAM directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            program=strdup(argv[1]);
        }else if(strcasecmp(directive, "PARAM")==0){
            if(!node){
                printf("PARAM directives are only valid in a NODE section, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            if(argc != 4){
                printf("malformed PARAM directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            thisparam = copy_param(argv[1], argv[2], argv[3]);
            ADD_LIST(thisparam, param_list);
        }else if(strcasecmp(directive, "END")==0){
            error_t result;
            
            if(!node){
                printf("You must begin a NODE section to end it, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }
            
            if(argc != 2 || strcasecmp(argv[1], "NODE")!=0){
                printf("Malformed END directive, line %d.\n", lineno);
                fclose(runf);
                return -1;
            }

            if(config == NULL || program == NULL){
                printf("NODE ending on line %d was invalid, missing CONFIG or PROGRAM.\n", lineno);
                free(config);
                free(program);
                fclose(runf);
                return -1;
            }

            node=0;
            result=atmega_load_chip(*chips, chipid, config, program);
            free(config);
            config=NULL;
            
            if(result < 0){
                int j;
                printf("Error loading chip %d.\n", chipid);
                
                /* failed on chip chipid, destroy all ones made prior */
                for(j=0; j<chipid; j++){
                    atmega_destroy_chip((*chips)+j);
                }
                
                fclose(runf);
                return -1;
            } else {
                char path[256];

                if(program[0] == '/'){
                    path[0]='\0';
                }else{
                    getcwd(path, sizeof(path));
                }

                printf( "Loaded file %s/%s for chip %d.\n", path, program, chipid );
                fflush( stdout ); 
            }

            thisparam = param_list;
            param_list = NULL;
            while(thisparam) {
                /* Send parameter to the appropriate device */
                thisdev = (*chips)[chipid]->device_head;
                while(thisdev) {
                    if(strcasecmp(thisdev->name, thisparam->device) == 0) {
                        ioman_param_func* paramfunc
                            = (ioman_param_func*)dlsym(thisdev->dlhandle, "set_param");
                        if(paramfunc == NULL) {
                            printf("Device %s does not have a set_param() function.\n", thisdev->name);
                            //TODO: Cleanup
                            abort();
                        }
                        (*paramfunc)(thisdev->dev, thisparam->key, thisparam->value);
                        break;
                    }
                }
                if(thisdev == NULL) {
                    // TODO: Board identification?  Cleanup
                    printf("Device %s not found.\n", thisdev->name);
                    abort();
                }

                /* Move to next parameter */
                nextparam = thisparam->next;
                free(thisparam);
                thisparam = nextparam;
            }

            free(program);
            program=NULL;

        }else{
            printf("unknown directive '%s', line %d\n", directive, lineno);
            fclose(runf);
            return -1;
        }
    }

    /* if there aren't as many chips as NCHIPS said there were */
    if(chipid + 1 < *nchips){
        *nchips = chipid + 1;
    }

    return 0;
}

static error_t atmega_parse_cpu_conf(atmega_chip *chip, char *token){
    char cpuconf[256];
    FILE *cpuf;
    int argc;
    char *argv[PARSER_MAX_REGS];
    int lineno=0;
    char *directive=NULL;

    io_reg_table=table_create(regcompare);

    snprintf(cpuconf, sizeof(cpuconf), "%s.conf", token);
    if((cpuf=open_config(cpuconf))==NULL){
        printf("Cannot open CPU configuration %s.\n", cpuconf);
        return -1;
    }

    while(parse_line(cpuf, &lineno, &argc, argv)){
        directive=argv[0];
        if(strcasecmp(directive, "REG")==0){
            char *reg, *addrs;
            io_reg *data;
            long addr;

            if(argc != 3){
                printf("invalid REG directive line %d.\n", lineno);
                fclose(cpuf);
                return -1;
            }

            reg=argv[1];
            addrs=argv[2];

            addr=strtol(addrs, NULL, 0);
            if(addr < 0 || addr >= PARSER_MAX_REGS){
                printf("Value 0x%lx out of range for register %s, line %d.\n", addr, reg, lineno);
                fclose(cpuf);
                return -1;
            }

            data=malloc(sizeof(*data));
            data->name=strdup(reg);
            data->reg=(uint8_t)addr;
            table_store(io_reg_table, data);
        }else if(strcasecmp(directive, "IO_OFFSET")==0){
            if(argc != 2){
                printf("Invalid IO_OFFSET directive, line %d.\n", lineno);
                fclose(cpuf);
                return -1;
            }
            chip->io_offset=strtol(argv[1], NULL, 0);
        }else if(strcasecmp(directive, "IO_LAST")==0){
            if(argc != 2){
                printf("Invalid IO_LAST directive, line %d.\n", lineno);
                fclose(cpuf);
                return -1;
            }
            chip->io_last=strtol(argv[1], NULL, 0);
        }else if(strcasecmp(directive, "SRAM_BYTES")==0){
            if(argc != 2){
                printf("Invalid SRAM_BYTES directive, line %d.\n", lineno);
                fclose(cpuf);
                return -1;
            }
            chip->sram_bytes=strtol(argv[1], NULL, 0);
        }else if(strcasecmp(directive, "FLASH_WORDS")==0){
            if(argc != 2){
                printf("Invalid FLASH_WORDS directive, line %d.\n", lineno);
                fclose(cpuf);
                return -1;
            }
            chip->flash_words=strtol(argv[1], NULL, 0);
        }else if(strcasecmp(directive, "PC_BITS")==0){
            if(argc != 2){
                printf("Invalid PC_BITS directive, line %d.\n", lineno);
                fclose(cpuf);
                return -1;
            }
            chip->pc_bits=strtol(argv[1], NULL, 0);
        }else if(strcasecmp(directive, "PC_MASK")==0){
            if(argc != 2){
                printf("Invalid IO_LAST directive, line %d.\n", lineno);
                fclose(cpuf);
                return -1;
            }
            chip->pc_mask=strtol(argv[1], NULL, 0);
        }else{
            printf("unknown directive '%s', line %d\n", directive, lineno);
            fclose(cpuf);
            return -1;
        }
    }
    fclose(cpuf);

    /* and then do some lookups to map magic registers */
    io_reg_table_lookup("SPL", &chip->spl_io);
    io_reg_table_lookup("SPH", &chip->sph_io);
    io_reg_table_lookup("SREG", &chip->sreg_io);
    io_reg_table_lookup("RAMPZ", &chip->rampz_io);
    io_reg_table_lookup("EIND", &chip->eind_io);

    return 0;
}

static void io_reg_table_destroy(){
    table_destroy(io_reg_table, regfree);
}

static error_t load_ioman_device(char *soname, device_funcs *df){
    char sopath[1024];
    int i=0;

    bzero(df, sizeof(*df));

    /* load the library */
    while(df->dlhandle==NULL && so_paths[i] != NULL){
        snprintf(sopath, sizeof(sopath), so_paths[i], soname);
        df->dlhandle=dlopen(sopath, RTLD_LAZY);
        i++;
    }
    if(df->dlhandle==NULL){
        printf("could not find device library '%s' in search path.\n", soname);
        return -1;
    }

    df->create_device=(ioman_create_func*)dlsym(df->dlhandle, "create_device");
    df->add_register=(ioman_register_func*)dlsym(df->dlhandle, "add_register");
    df->add_interrupt=(ioman_interrupt_func*)dlsym(df->dlhandle, "add_interrupt");
    df->add_clock=(ioman_clock_func*)dlsym(df->dlhandle, "add_clock");
    df->set_param=(ioman_param_func*)dlsym(df->dlhandle, "set_param");
    df->set_device=(ioman_uses_func*)dlsym(df->dlhandle, "set_device");

    return 0;
}

error_t atmega_parse_config(atmega_chip *chip, char *config){
    FILE *configf;
    char *directive=NULL;
    device_funcs df;
    device_obj *dev=NULL;

    int argc;
    char *argv[PARSER_MAX_ARGS];
    char buff[1024];

    int device=0;
    char devname[256];
    int lineno=0;

    line_buff=buff;
    
    /* init structures */
    bzero(chip->register_head, sizeof(chip->register_head));
    chip->interrupt_head=NULL;
    chip->device_head=NULL;
    chip->clock_head=NULL;

    /* parse config file */
    if((configf=open_config(config))==NULL){
        return -errno;
    }
    while(parse_line(configf, &lineno, &argc, argv)){
        directive=argv[0];
        if(strcasecmp(directive, "CPU")==0){
            if(argc==2){
                if(atmega_parse_cpu_conf(chip, argv[1]) < 0){
                    printf("Error loading CPU config, line %d.\n", lineno);
                    io_reg_table_destroy();
                    fclose(configf);
                    return -1;
                }
            }else{
                printf("invalid CPU directive, line %d\n", lineno);
                fclose(configf);
                return -1;
            }
        }else if(strcasecmp(directive, "DEVICE")==0){
            if(device){
                printf("you must END DEVICE for %s before beginning a new one, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }else{
                char *soname, *objname;
                if(argc != 3){
                    printf("invalid DEVICE directive, line %d.\n", lineno);
                    fclose(configf);
                    return -1;
                }
                soname=argv[1];
                objname=argv[2];
                strncpy(devname, objname, sizeof(devname));

                if(load_ioman_device(soname, &df) < 0){
                    fclose(configf);
                    return -1;
                }

                if(df.create_device == NULL){
                    printf("mandatory function 'create_device' not implemented in device %s, line %d.\n", devname, lineno);
                    fclose(configf);
                    return -1;
                }

                dev=(*df.create_device)(objname);
                if(dev==NULL){
                    printf("Error creating device %s, line %d.\n", devname, lineno);
                    fclose(configf);
                    return -1;
                }

                dev->dlhandle=df.dlhandle;
                dev->name = strdup(objname);
                dev->type = strdup(soname);
                dev->chip = chip;
                if(dev->name == NULL) {
                    printf("Error allocating name for device %s, line %d.\n", devname, lineno);
                    fclose(configf);
                    return -1;
                }
                dev->device_event=ioman_device_event;
                dev->device_debug=ioman_device_debug;
                dev->add_clock = device_add_clock;
                dev->remove_clock = device_remove_clock;

                ADD_LIST(dev, chip->device_head);

                device=1;
            }
        }else if(strcasecmp(directive, "CONNECT")==0){
            char *cpureg, *devreg, *smask;
            unsigned int bmask_l;
            uint8_t bmask, cpuregnum;
            register_obj *reg;

            if(!device){
                printf("CONNECT only valid inside a DEVICE block, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }

            if(argc != 4){
                printf("Malformed CONNECT directive, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }

            cpureg=argv[1];
            devreg=argv[2];
            smask=argv[3];

            bmask_l=strtol(smask, NULL, 0);
            bmask=bmask_l;

            if(io_reg_table_lookup(cpureg, &cpuregnum) < 0){
                printf("invalid cpu register '%s'., line %d\n", cpureg, lineno);
                fclose(configf);
                return -1;
            }

            if(df.add_register==NULL){
                printf("CONNECT directive used, but add_register unimplemented by device %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            if((reg=(*df.add_register)(dev, devreg)) == NULL){
                printf("error executing add_register for device %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            reg->mask=bmask;
            reg->maskshift=0;
            while(((reg->mask >> reg->maskshift) & 0x01) == 0){
                reg->maskshift++;
            }

            ADD_LIST(reg, chip->register_head[cpuregnum]);
        }else if(strcasecmp(directive, "INT")==0){
            interrupt_obj *interrupt;
            char *name, *svector;
            uint8_t vector;

            if(!device){
                printf("INT only valid inside a DEVICE block, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }

            if(argc != 3){
                printf("Malformed INT directive, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }
            name=argv[1];
            svector=argv[2];

            if(df.add_interrupt==NULL){
                printf("INT directive used, but add_interrupt unimplemented by device %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            
            if((interrupt=(*df.add_interrupt)(dev, name)) == NULL){
                printf("error executing add_interrupt in %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            vector=strtol(svector, NULL, 0);
            interrupt->vector=vector;
            if(vector > 63) {
                printf("interrupt number too high in %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }

            interrupt->vector_mask = (uint64_t)1 << vector;
            interrupt->chip = chip;
            interrupt->set_interrupt = set_interrupt;
            interrupt->clear_interrupt = clear_interrupt;

            ADD_LIST(interrupt, chip->interrupt_head);
        }else if(strcasecmp(directive, "CLOCKED")==0){
            clock_obj *clock;

            if(!device){
                printf("CLOCK only valid inside a DEVICE block, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }
            
            if(argc != 1){
                printf("Malformed CLOCKED directive, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }

            if(df.add_clock==NULL){
                printf("CLOCK directive used, but add_clock unimplemented by %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            if((clock=(*df.add_clock)(dev)) == NULL){
                printf("error executing add_clock in %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }

            ADD_LIST(clock, chip->clock_head);
        }else if(strcasecmp(directive, "PARAM")==0){
            char *key, *value;

            if(!device){
                printf("PARAM only valid inside a DEVICE block, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }

            if(argc != 3){
                printf("Malformed PARAM directive in device %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            key=argv[1];
            value=argv[2];

            if(df.set_param==NULL){
                printf("PARAM directive used, but set_param unimplemented by %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            (*df.set_param)(dev, key, value);
        }else if(strcasecmp(directive, "USES")==0){
            device_obj *curr;

            if(!device){
                printf("USES only valid inside a DEVICE block, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }

            if(argc != 3){
                printf("Malformed USES directive in device %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }

            for(curr = chip->device_head;
                    (curr != NULL) && (strcasecmp(curr->name, argv[2]) != 0);
                    curr = curr->next);

            if(curr == NULL) {
                printf("USES references a non-existant device %s in device %s, line %d.\n", argv[2], devname, lineno);
                fclose(configf);
                return -1;
            }

            if(df.set_device == NULL) {
                printf("USES directive used, but set_device unimplemented by %s, line %d.\n", devname, lineno);
                fclose(configf);
                return -1;
            }
            (*df.set_device)(dev, argv[1], curr);
        }else if(strcasecmp(directive, "END")==0){
            if(!device){
                printf("You must begin a DEVICE section to end it, line %d.\n", lineno);
                fclose(configf);
                return -1;
            }else{
                if(argc != 2 || strcasecmp(argv[1], "DEVICE")!=0){
                    printf("Malformed END directive in device %s, line %d.\n", devname, lineno);
                    fclose(configf);
                    return -1;
                }

                device=0;
            }
        }else{
            printf("unknown directive '%s', line %d\n", directive, lineno);
            fclose(configf);
            return -1;
        }
    }

    fclose(configf);
    io_reg_table_destroy();

    return 0;
}

