/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>
#include <libelf.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <gtk/gtk.h>

#include "atemu.h"
#include "breakpoints.h"
#include "../os.h"
#include "table.h"
#include "xatdb_elf.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "func_data.h"
#include "type_data.h"
#include "stabs.h"

#define FNAME_STACK_SIZE 256

struct nlist {
    unsigned long n_strx;         /* index into string table of name */
    unsigned char n_type;         /* type of symbol */
    unsigned char n_other;        /* misc info (usually empty) */
    unsigned short n_desc;        /* description field */
    unsigned long n_value;        /* often an address */
};

/* byte-swaps a long */
uint32_t swap_l(uint32_t x){
    return (((x & 0x000000ff) << 24) |
            ((x & 0x0000ff00) << 8) |
            ((x & 0x00ff0000) >> 8) |
            ((x & 0xff000000) >> 24));
}

/* byte-swaps a short */
uint16_t swap_s(uint16_t x){
    return (((x & 0x00ff) << 8) |
            ((x & 0xff00) >> 8));
}

/* converts an nlist entry of byte order order to local machine endianess */
void convert_nlist(struct nlist *nl, unsigned order){
    if(ELF_BYTEORDER != order){
        /* only need to byteswap if byteorder differs */
        int i;
        
        nl[0].n_strx = swap_l(nl[0].n_strx);
        nl[0].n_desc = swap_s(nl[0].n_desc);
        nl[0].n_value = swap_l(nl[0].n_value);

        for(i=1; i<=nl[0].n_desc; i++){
            nl[i].n_strx = swap_l(nl[i].n_strx);
            nl[i].n_desc = swap_s(nl[i].n_desc);
            nl[i].n_value = swap_l(nl[i].n_value);
        }
    }
}

void create_type( int chipid, char *name, char *types )
{
    char new_desc;
    char *new_type;
    char *curr, *semi, *comma;
    char type_desc;
    char *type_def;

    struct type_data *td;
    struct type_data *ref;
    struct field *f;

    /* For determining byte size of ranges. */
    long long int begin, end;
    long long unsigned int range;
    int bits, bytes, size = 0, align = 0;

    new_desc = *types;

    /* CHECK: default? */
    bytes = 4;

    switch( new_desc ) {
        case SD_TYPE_NAME:
            new_type = types + 1;
            printf( "Creating type: <%s>\n", name );
            break;

        case SD_STRUCT:
            new_type = types + 1;
            printf( "Creating struct: <%s>\n", name );
            break;

        default:
            /* guess that this is the type label anyway */
            new_type = types + 1;
            printf( "Unknown type to create.  Done.\n" );
            return;
    }

    curr = strchr( types, '=' );

    /* Really should be there... */
    if ( curr == NULL )
    {
        printf( "create_type: '=' not there!\n" );
    }
    *curr = '\0';

    curr++;
    type_desc = *curr;

        /* Attributes first... like size or alignment */
    if ( type_desc == TD_TYPE_ATTR )
    {
        /* Point curr to char. after '@' (TD_TYPE_ATTR) */
        curr++;

        if ( *curr == 'P' ){
            printf( "\tPacked.\n" );
            curr++;
        } else if ( *curr == 'S' ) {              
            printf( "\tString.\n" );
            curr++;
        } else if ( *curr == 'p' ) {
            printf( "\tPointer.\n" );
            strtod( (curr + 1), &curr );
        } else if ( *curr == 'a' ) {
            align = (int) strtod( (curr + 1), &curr );
            printf( "\tAlignement: %d\n", align );
        } else if ( *curr == 's' ) {
            size = (int) strtod( (curr + 1), &curr );
            printf( "\tSize: %d\n", size );
        }

        /* Get past the ";" */
        curr++;
        type_desc = *curr;
    }

    switch( type_desc ) {
        /* A simple reference? */
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case TD_TYPE_REF:
            /* CHECK: Write code for reference... */
            type_def = curr;
            printf( "\tReference. %s -> %s\n", new_type, type_def );

            ref = id_to_type_data( chipid, type_def );

            if ( ref != NULL )
            {
                td = create_type_data( name, new_type, ref->byte_length,
                    ref->begin, ref->end );

                if ( td == NULL )
                {
                    printf("Couldn't create type_data.\n" ); 
                    return;
                }

                add_type_data( chipid, td );
            }

            break;

        case TD_RANGE:
            type_def = (curr + 1);
            semi = strchr( type_def, ';' );
            if ( semi == NULL )
            {
                printf( "create_type: TD_RANGE should have ';'!\n" );
                break;
            }
            *semi = '\0';
            
            begin = strtoll( (semi + 1), &semi, 0 );
            end = strtoll( (semi + 1), &semi, 0 );

            if (( begin > 0 ) && ( end == 0 ))
            {
                bytes = begin;
                bits = bytes * 8;
            } else {
                if (( begin > 0 ) && ( end > 0 ))
                {
                    begin = -begin;
                }

                range = end - begin;

                /* A poor man's lg function... */
                bits = 0;
                while( range )
                {
                    bits++;
                    range >>= 1;
                }
                bytes = (int) ceil( (double) bits / 8 );
            }

            printf( "\tRange.\n" );
            printf( "\tDef: %s -> %s\n", new_type, type_def );
            printf( "\tSize: %d (%d)\n", bytes, bits );

            td = create_type_data( name, new_type, bytes, begin, end );
            if ( td == NULL )
            {
                printf("Couldn't create type_data.\n" ); 
                return;
            }

            add_type_data( chipid, td );
            break;

        case TD_STRUCT:
            size = strtod( (curr + 1), &curr );
            printf( "\tStructure:\n" );
            printf( "\tDef: %s\n", new_type );
            printf( "\tSize: %d\n", size );

            td = create_type_data( name, new_type, size, 0, 0 );
            if ( td == NULL )
            {
                printf("Couldn't create type_data.\n" ); 
                return;
            }

            /* Note: This _is_ bad parsing, but there are reasons...
                The delimiters can actually be inside the fields... 
                So... I've decided I don't care what type these fields
                are.  I only want the size and name.  If anyone feels
                so inclined to fix this beast, feel free.
                    -- DMB
             */
            while( *(curr) != ';' )
            {
                semi = strchr( curr, ';' );
                if ( semi == NULL )
                {
                    printf( "Ill-formed struct. ';' is missing\n" );
                    free_type_data( (void *) td );
                    return;
                }
                *semi = '\0';

                /* semi will now point to first ':' ... bad naming */
                semi = strchr( curr, ':' );
                *semi = '\0';
                name = curr;

                curr = semi + 1;

                comma = strrchr( curr, ',' );
                size = strtod( (comma + 1), &curr );
                
                f = create_field( name, size ); 
                add_field_to_type( td, f );

                printf( "\t\tField %s<%d>[%c] added.\n", name, size, *curr );

                curr++;
            }

            add_type_data( chipid, td );
            break;

        case TD_FLOAT_BUILT_IN:
            
            break;

        default:
            printf( "create_type: couldn't find type_desc (%c)\n",
                type_desc );
            break;
    }
}

/* this parses the stab list you pass it, storing things in nice structs */
void build_maps( char * strings, struct nlist * nl, int chipid ){
    /* there are n_desc MORE stab entries after this one */
    int maxstabs=nl[0].n_desc;
    int i;
    unsigned long func_addr = 0;
    unsigned char in_func = 0;
    char func_name[128];
    unsigned char in_c_line = 0;
    unsigned long old_addr = 0;
    unsigned long curr_line = 0;
    struct func_data *fd = NULL;
    struct local *loc = NULL;
    struct c_line *cl = NULL;
    int len;
    int curr_id;

    char * colon;
    char local_name[128];
    char type_name[128];
    int is_param;
    int type;

    char *fname_path = "";
    char *fname_stack[FNAME_STACK_SIZE];
    int fname_index=0;
    
    for(i=0; i < FNAME_STACK_SIZE; i++){
        fname_stack[i] = NULL;
    }

    for(i=1; i <= maxstabs; i++){
        unsigned long off = nl[i].n_strx;
        switch(nl[i].n_type){
            case N_FUN:
                /* Function */
                if ( !in_func ) {
                    /* start of a function */
                    func_addr = nl[i].n_value;

                    strncpy( func_name, (char *) (strings + off), 128 );
                    func_name[127] = 0;
                    if ( ( colon = strchr( func_name, ':' ) ) )
                        *colon = 0;

                    fd = create_func_data( func_name, func_addr, func_addr );

                    in_func = 1;
                } else {
                    /* end of last function started */

                    /* end any line mapping in progress */
                    if ( in_c_line ) {
                        in_c_line = 0;

                        add_addr_file_mapping( chipid, old_addr,
                                nl[i].n_value + func_addr, curr_line, fname_stack[fname_index] );
                    }

                    if ( fd != NULL )
                    {
                        fd->addr_end = func_addr + nl[i].n_value;

                        add_func_data( chipid, fd );
                    }

                    in_func = 0;
                    curr_id = 0;
                }
                break;

            case N_GSYM:
                printf( "Global: [%s]\n", (strings + off) );
                strncpy( local_name, (char *) (strings + off), 128 );
                local_name[127] = 0;
                if ( ( colon = strchr( local_name, ':' ) ) )
                    *(colon++) = 0;

                if ( *colon == 'G' )
                {
                    struct type_data *td;

                    colon++;

                    strncpy( type_name, colon, 128 );
                    type_name[127] = 0;

                    td = id_to_type_data( chipid, type_name );

                    size_global( chipid, local_name, td );
                }

                break;

            case N_RSYM:
            case N_LSYM:
            case N_PSYM:
                /* Locals and Params */

                strncpy( local_name, (char *) (strings + off), 128 );
                local_name[127] = 0;
                if ( ( colon = strchr( local_name, ':' ) ) )
                    *(colon++) = 0;

                if ( in_func ) 
                {
                    struct type_data *td;

                    if ( fd != NULL )
                    {
                        if ( nl[i].n_type == 0x40 ) {
                            if ( *colon == 'P' ) {
                                is_param = 1;
                            } else if ( *colon == 'r' ) {
                                is_param = 0;
                            }

                            colon++;

                            type = LOCAL_TYPE_REG;
                        } else if ( nl[i].n_type == 0xa0 ) {
                            if ( *colon == 'p' ) {
                                is_param = 1;
                            } else {
                                is_param = 0;
                            }

                            colon++;

                            type = LOCAL_TYPE_STACK;
                        } else {
                            is_param = 0;

                            type = LOCAL_TYPE_STACK;
                        }

                        td = id_to_type_data( chipid, colon );

                        loc = create_local( local_name, is_param, type,
                            nl[i].n_value, td, curr_id++ );

                        add_local_to_func( fd, loc );
                    }
                } else if ( nl[i].n_type == N_LSYM ) {
                    if ( *colon != '(' )
                    {
                        create_type( chipid, local_name, colon );
                    }
                }
                break;
            case N_EINCL:
                /* end include file */
                /* this one is stack-based */

                free(fname_stack[fname_index]);
                
                fname_index--;
                if(fname_index < 0){
                    printf("decremented stack below 0, oh no!\n");
                    fname_index = 0;
                }
                break;
            case N_BINCL:
                /* begin include file */
                /* this one is stack-based */

                fname_index++;
                if(fname_index >= FNAME_STACK_SIZE){
                    printf("incremented stack above %d, oh no!\n", FNAME_STACK_SIZE);
                    fname_index = FNAME_STACK_SIZE - 1;
                }
                if((strings + off)[0] != '/'){
                    fname_stack[fname_index] = malloc(strlen(fname_path) + strlen(strings + off) + 1);
                    // non 'n' versions ok, as lengths carefully calculated
                    strcpy(fname_stack[fname_index], fname_path);
                    strcat(fname_stack[fname_index], strings + off);
                }else{
                    fname_stack[fname_index] = strdup(strings + off);
                }
                add_cfile(chipid, fname_stack[fname_index]);
                break;        
            case N_SOL:
            case N_SO:
                /* start of C/include file */

                /* end any line mapping in progress */
                if ( in_c_line ) {
                    in_c_line = 0;

                    add_addr_file_mapping( chipid, old_addr,
                            nl[i].n_value, curr_line, fname_stack[fname_index] );
                }

                len = strlen(strings + off);
                /* last character of (strings + off) is a '/', this is a path */
                if( len > 0 && (strings + off)[len - 1] == '/'){
                    fname_path = strings + off;
                }else{
                    /* it's the actual file name */
                    free(fname_stack[fname_index]);
                    if(len > 0 && (strings + off)[0] != '/'){
                        fname_stack[fname_index] = malloc(strlen(fname_path) + strlen(strings + off) + 1);
                        // non 'n' versions ok, as lengths carefully calculated
                        strcpy(fname_stack[fname_index], fname_path);
                        strcat(fname_stack[fname_index], strings + off);
                    }else{
                        fname_stack[fname_index] = strdup(strings + off);
                    }
                    add_cfile(chipid, fname_stack[fname_index]);
                }
                break;
            case N_SLINE:
                /* Line Number */
                if ( in_c_line ) {
                    if ( in_func ) {
                        add_addr_file_mapping( chipid, old_addr,
                                nl[i].n_value + func_addr, curr_line, fname_stack[fname_index] );
                    } else {
                        add_addr_file_mapping( chipid, old_addr,
                                nl[i].n_value, curr_line, fname_stack[fname_index] );
                    }
                }

                if ( in_func ) {
                    curr_line = nl[i].n_desc;
                    old_addr = nl[i].n_value + func_addr;

                    cl = create_c_line( old_addr, curr_line );

                    add_c_line_to_func( fd, cl );
                } else {
                    curr_line = nl[i].n_desc;
                    old_addr = nl[i].n_value;
                }

                in_c_line = 1;
                break;
            default:
//                printf("Unhandled stab: type 0x%02x\n",nl[i].n_type);
                break;
        }
        // prevent the UI from becoming unresponsive
        g_main_context_iteration(NULL, FALSE);
        if(xatdb_quit){
            return;
        }
    }

    /* free any strings left in the stack */
    for(i=0; i <= fname_index; i++){
        free(fname_stack[i]);
    }

    /* make the C stepping breakpoints */
    generate_auto_breakpoints(chipid);

    fflush(stdout);
}

int parse_elf(const char * elfname, int chipid ){
    int elffile;
    Elf *elf;
    Elf32_Shdr *shdr;
    Elf32_Ehdr *ehdr;
    Elf_Scn *scn, *stab=NULL, *stabstr=NULL, 
        *symtab = NULL, *strtab = NULL;
    Elf_Data *data;

    int cnt;
    struct nlist *nl;
    char *strings;

    if(xatdb_quit){
        return -1;
    }

    elffile=open(elfname, O_RDONLY);
    if(elffile < 0){
        printf("Error opening %s!\n", elfname);
        return -1;
    }

    /* set version of elf */
    elf_version(EV_CURRENT);

    /* open the elf file */
    elf=elf_begin(elffile, ELF_C_READ, NULL);
    if(elf == NULL){
        printf("%s: %s.\n", elfname, elf_errmsg(elf_errno()));
        close(elffile);
        return -1;
    }

    /* get some sections */
    if (((ehdr = elf32_getehdr(elf)) == NULL) ||
        ((scn = elf_getscn(elf, ehdr->e_shstrndx)) == NULL) ||
        ((data = elf_getdata(scn, NULL)) == NULL)){
        printf("%s: %s.\n", elfname, elf_errmsg(elf_errno()));
        elf_end(elf);
        close(elffile);
        return -1;
    }

    /* print names of sections */
    for (cnt = 1, scn = NULL; (scn = elf_nextscn(elf, scn)); cnt++) {
        if ((shdr = elf32_getshdr(scn)) == NULL){
            printf("%s: %s.\n", elfname, elf_errmsg(elf_errno()));
            elf_end(elf);
            close(elffile);
            return -1;
        }
        if(strcmp((char *)data->d_buf + shdr->sh_name, ".stab") == 0){
            stab = scn;
        }else if(strcmp((char *)data->d_buf + shdr->sh_name, ".stabstr") == 0){
            stabstr = scn;
        }else if(strcmp((char *)data->d_buf + shdr->sh_name, ".strtab") == 0){
            strtab = scn;
        }else if(strcmp((char *)data->d_buf + shdr->sh_name, ".symtab") == 0){
            symtab = scn;
        }
    }

    /* do the symbols */
    if ( (strtab == NULL) || (symtab == NULL) )
    {
        printf("Missing either .symtab or .strtab.  Ergo, no symbol info.\n");
    } else {
        /* borrowed from: 
           http://developers.sun.com/solaris/articles/elf.html */

        Elf32_Sym *esym, *lastsym;
        char *name;

        data = elf_getdata(strtab, NULL);
        strings = (char *) data->d_buf;
        data = elf_getdata(symtab, NULL);
        shdr = elf32_getshdr(symtab);

        esym = (Elf32_Sym*) data->d_buf;
        lastsym = (Elf32_Sym*) ((char *)data->d_buf + data->d_size);

        /* now loop through the symbol table */
        for (; esym < lastsym; esym++)
        {
            if ((esym->st_value == 0) ||
                (ELF32_ST_BIND(esym->st_info)== STB_WEAK) ||
                (ELF32_ST_BIND(esym->st_info)== STB_NUM) ||
                (ELF32_ST_TYPE(esym->st_info)!= STT_OBJECT)) {
                continue;
            }

            name = elf_strptr(elf,shdr->sh_link, (size_t)esym->st_name);

            /* add to global table */
            if( name != NULL )
            {
                printf("%s : <%08lx>\n", name, 
                    (long unsigned int) esym->st_value & (~0x800000) );

                add_global( chipid, name, esym->st_value & (~0x800000) );
            }
            // prevent the UI from becoming unresponsive
            g_main_context_iteration(NULL, FALSE);
            if(xatdb_quit){
                elf_end(elf);
                close(elffile);
                return -1;
            }
        }
    }

    if(stab == NULL || stabstr == NULL){
        printf("Missing either .stab or .stabstr.  Ergo, no debug info.\n");
        elf_end(elf);
        close(elffile);
        return -1;
    }

    /* get .stab data section (fixing endianness as we go) */
    data = elf_getdata(stab, NULL);
    nl=(struct nlist*)data->d_buf;
    convert_nlist(nl, ehdr->e_ident[EI_DATA]);

    /* get the stab string table data section */
    data = elf_getdata(stabstr, NULL);
    strings=(char *)data->d_buf;

    build_maps( strings, nl, chipid );
    if(xatdb_quit){
        elf_end(elf);
        close(elffile);
        return -1;
    }

    /* do the cfile sort thing */
    make_cfile_menu(chipid);

    elf_end(elf);
    close(elffile);
    return 0;
}
